@@ -16,7 +16,7 @@ package parser
 import (
 	"testing"
 
-	"github.com/prometheus/prometheus/util/testutil"
+	"github.com/stretchr/testify/require"
 )
 
 type testCase struct {
@@ -276,6 +276,9 @@ var tests = []struct {
 			}, {
 				input:    `unless`,
 				expected: []Item{{LUNLESS, 0, `unless`}},
+			}, {
+				input:    `@`,
+				expected: []Item{{AT, 0, `@`}},
 			},
 		},
 	},
@@ -337,6 +340,23 @@ var tests = []struct {
 				input:    "bool",
 				expected: []Item{{BOOL, 0, "bool"}},
 			},
+			{
+				input:    "atan2",
+				expected: []Item{{ATAN2, 0, "atan2"}},
+			},
+		},
+	},
+	{
+		name: "preprocessors",
+		tests: []testCase{
+			{
+				input:    `start`,
+				expected: []Item{{START, 0, `start`}},
+			},
+			{
+				input:    `end`,
+				expected: []Item{{END, 0, `end`}},
+			},
 		},
 	},
 	{
@@ -721,20 +741,20 @@ func TestLexer(t *testing.T) {
 					}
 					if !hasError {
 						t.Logf("%d: input %q", i, test.input)
-						t.Fatalf("expected lexing error but did not fail")
+						require.Fail(t, "expected lexing error but did not fail")
 					}
 					continue
 				}
 				if lastItem.Typ == ERROR {
 					t.Logf("%d: input %q", i, test.input)
-					t.Fatalf("unexpected lexing error at position %d: %s", lastItem.Pos, lastItem)
+					require.Fail(t, "unexpected lexing error at position %d: %s", lastItem.Pos, lastItem)
 				}
 
 				eofItem := Item{EOF, Pos(len(test.input)), ""}
-				testutil.Equals(t, lastItem, eofItem, "%d: input %q", i, test.input)
+				require.Equal(t, lastItem, eofItem, "%d: input %q", i, test.input)
 
 				out = out[:len(out)-1]
-				testutil.Equals(t, test.expected, out, "%d: input %q", i, test.input)
+				require.Equal(t, test.expected, out, "%d: input %q", i, test.input)
 			}
 		})
 	}
